/**
 * Title:        CoCompose
 * Copyright:    Copyright (c) 2002, 2003
 * Author:       Shawn Lorae Stutzman
 * Author:       Dennis Wagelaar
 * Organisation: University of Twente, Vrije Universiteit Brussel
 * Description:  CoCompose Design Language Tool
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License, version 2 as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  (See the file "COPYING" that is included with this source distribution.)
 *
 *  Taken from:
 *  File:  Spring.java  (Implementation of Kamada and Kawai's spring algorithm
 *                      with modifications).
 *
 *  12/12/96    Shawn Lorae Stutzman
 */

package EDU.utwente.CoCompose.algorithm;

import java.lang.reflect.*;
import java.util.*;
import EDU.auburn.VGJ.algorithm.*;
import EDU.auburn.VGJ.algorithm.shawn.Queue;
import EDU.auburn.VGJ.graph.*;
import EDU.auburn.VGJ.util.*;
import EDU.utwente.CoCompose.lang.*;
import EDU.utwente.CoCompose.util.CoComposeException;

/**
 * Model layout algorithm adapted from the EDU.auburn.VGJ.algorithm.shawn.Spring
 * algorithm to cope with CoCompose models. Class to implement Kamada and
 * Kawai's spring algorithm with modifications.
 *
 * @author Dennis Wagelaar
 * @author Shawn Lorae Stutzman
 * @version 0.1.10, 22 July 2003
 * @see EDU.auburn.VGJ.algorithm.shawn.Spring
 */

public class ModelSpring implements GraphAlgorithm, Runnable
{
  //////////////////////////////////////////////////////////////////////////////
  // Constants
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Constant to limit number of times thru "inner" loop.
   */
  private static final int MAX_TIMES_REPOSITIONED = 10;

  //////////////////////////////////////////////////////////////////////////////
  // Attributes
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Thread that runs algorithm passes.
   */
  private Thread thread;
  private DRect rect;
  private double xmax, xmin, ymax, ymin;
  /**
   * This is used to enable a quick lookup of nodes' array values (which might
   * differ from their index values).
   */
  private int orderedList[];
  /**
   * Nodes in graph; will be set in compute function.
   */
  private int N;
  /**
   * Edges in graph; set in compute function.
   */
  private int EDGES;
  /**
   * The graphical distances between all pairs.
   */
  private long D[][];
  /**
   * Spring strengths.
   */
  private double K[][];
  /**
   * Spring constant.
   */
  private double Ko;
  /**
   * Ideal spring lengths.
   */
  private double L[][];
  private double epsilon;
  private double delta[];
  /**
   * Counts number of times thru "inner" loop.
   */
  private int numTimesRepositioned = 0;
  /**
   * Counts number of times each vertex has been selected to move.
   */
  private int NUM_TIMES_MOVED[];
  /**
   * These variables control the E stopping condition and how it keeps track of
   * its information.
   */
  private int HY_SIZE = 10;
  private double E, E_HY[];
  private int COUNTER = 0;
  /**
   * The node chosen to move at each iteration.
   */
  private Node mv = null;
  private double partial_x, partial_y, partial_xx, partial_xy,
    partial_yx, partial_yy;
  private boolean connected;
  /**
   * The model to run the algorithm on
   */
  private Model model;
  /**
   * The graph updateable display.
   */
  private GraphUpdate display;
  /**
   * If true, algorithm should be running, otherwise it should stop at the first
   * possible occasion.
   */
  private boolean running = false;

  //////////////////////////////////////////////////////////////////////////////
  // Constructors
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Constructs a ModelSpring.
   */
  public ModelSpring()
  {
  } //ModelSpring

  //////////////////////////////////////////////////////////////////////////////
  // Private methods
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Initializes graph for computation.
   *
   * @param M The model to initialize.
   * @throws Exception if model is not valid for this algorithm.
   */
  private void init(Model M) throws Exception
  {
    Node v, u;
    int i;

    N = M.getFreeNodeCount();
    makeOrderedList(M);
    NUM_TIMES_MOVED = new int[N];
    for (i = 0; i < N; i++)
      NUM_TIMES_MOVED[i] = 0;
    EDGES=0;
    for (v = M.firstFreeNode(); v != null; v = M.nextFreeNode(v))
    {
      for (u = M.nextFreeNode(v); u != null; u = M.nextFreeNode(u))
        if (v.hasChild(u)) EDGES++;
    }
    epsilon = 0.5 * (N + EDGES);
    findDistances(M);
    find_l_and_k(M);
    delta = new double[N];
    E_HY = new double[HY_SIZE];
  } //init

  /**
   * Finds and sets distances between nodes.
   *
   * @param M The model to work on.
   */
  private void findDistances(Model M)
  {
    Node s, u, v, p, role;
    int i, j, vi, pi, si, ui;
    boolean done[];
    Queue queue;

    queue = new Queue();
    done = new boolean[M.getFreeNodeCount()];
    D = new long[M.getFreeNodeCount()][M.getFreeNodeCount()];
    /* Initialize the distance matrix to 0s */
    for (i = 0; i < M.getFreeNodeCount(); i++)
    {
      for (j = i; j < M.getFreeNodeCount(); j++)
      {
        D[i][j] = 0;
        D[j][i] = 0;
      }
    }
    /* For each node, do Dijkstra... */
    for (s = M.firstFreeNode(); s != null; s = M.nextFreeNode(s))
    {
      /* Initialize the done array to false */
      for (i = 0; i < M.getFreeNodeCount(); i++)
        done[i] = false;

      si = enum(M.getIndexFromNode(s));
      done[si] = true;
      for (v = M.firstFreeNode(); v != null; v = M.nextFreeNode(v))
      {
        /* find connections to free nodes. */
        if (s.hasChild(v))
        {
          queue.push(M.getIndexFromNode(v));
          queue.push(M.getIndexFromNode(s));
          vi = enum(M.getIndexFromNode(v));
          done[vi] = true;
        }
        // if s is a composite, count connections to roles as well
        if (s instanceof EDU.utwente.CoCompose.lang.Composite)
        {
          Enumeration roles = ((Composite) s).getRoles();
          while (roles.hasMoreElements())
          {
            role = (Node) roles.nextElement();
            if (role.hasChild(v))
            {
              queue.push(M.getIndexFromNode(v));
              queue.push(M.getIndexFromNode(s));
              vi = enum(M.getIndexFromNode(v));
              done[vi] = true;
            }
          }
        } //if composite
        // if v is a composite, count connections to roles as well
        if (v instanceof EDU.utwente.CoCompose.lang.Composite)
        {
          Enumeration roles = ((Composite) v).getRoles();
          while (roles.hasMoreElements())
          {
            role = (Node) roles.nextElement();
            if (s.hasChild(role))
            {
              queue.push(M.getIndexFromNode(v));
              queue.push(M.getIndexFromNode(s));
              vi = enum(M.getIndexFromNode(v));
              done[vi] = true;
            }
          }
        } //if composite
      } //for (v..
      while (!(queue.isEmpty()))
      {
        v = M.getNodeFromIndex(queue.pop());
        p = M.getNodeFromIndex(queue.pop());
        pi = enum(M.getIndexFromNode(p));
        vi = enum(M.getIndexFromNode(v));
        D[si][vi] = D[pi][si] + 1;
        D[vi][si] = D[si][vi];
        for (u = M.firstFreeNode(); u != null; u = M.nextFreeNode(u))
        {
          /* find connections to free nodes. */
          if (v.hasChild(u))
          {
            ui = enum(M.getIndexFromNode(u));
            if (!(done[ui]))
            {
              queue.push(M.getIndexFromNode(u));
              queue.push(M.getIndexFromNode(v));
              done[ui] = true;
            }
          }
          // if v is a composite, count connections to roles as well
          if (v instanceof EDU.utwente.CoCompose.lang.Composite)
          {
            Enumeration roles = ((Composite) v).getRoles();
            while (roles.hasMoreElements())
            {
              role = (Node) roles.nextElement();
              if (role.hasChild(u))
              {
                ui = enum(M.getIndexFromNode(u));
                if (!(done[ui]))
                {
                  queue.push(M.getIndexFromNode(u));
                  queue.push(M.getIndexFromNode(v));
                  done[ui] = true;
                }
              }
            }
          } //if composite
          // if u is a composite, count connections to roles as well
          if (u instanceof EDU.utwente.CoCompose.lang.Composite)
          {
            Enumeration roles = ((Composite) u).getRoles();
            while (roles.hasMoreElements())
            {
              role = (Node) roles.nextElement();
              if (v.hasChild(role))
              {
                ui = enum(M.getIndexFromNode(u));
                if (!(done[ui]))
                {
                  queue.push(M.getIndexFromNode(u));
                  queue.push(M.getIndexFromNode(v));
                  done[ui] = true;
                }
              }
            }
          } //if composite
        } //for (u..
      } //while
    } //for (s..
    /* Change the appropriate remaining 0's to infinities and set value of the
     * boolean variable connected */
    connected = true;
    for (i = 0; i < M.getFreeNodeCount(); i++)
    {
      for (j = i + 1; j < M.getFreeNodeCount(); j++)
      {
        if (D[i][j] == 0)
        {
          connected = false;
          D[i][j] = Long.MAX_VALUE;
          D[j][i] = Long.MAX_VALUE;
        }
      }
    }
  }//findDistances

  /**
   * Makes ordered list of all nodes that aren't CompositeRoles.
   *
   * @param M The model to work on.
   */
  private void makeOrderedList(Model M)
  {
    int highestIndex, i;
    Node v;

    orderedList = new int[M.getFreeNodeCount()];
    for (i = 0, v = M.firstFreeNode(); v != null; i++, v = M.nextFreeNode(v))
    {
      orderedList[i] = M.getIndexFromNode(v);
    }
  } //makeOrderedList

  /**
   * Retrieves the ordered list index of the node with the given graph index.
   *
   * @param Index The graph index of the node.
   * @return The ordered list index or -1 if not found.
   */
  private int enum(int Index)
  {
    int i;

    for (i = 0; i < Array.getLength(orderedList); i++)
    {
      if (orderedList[i] == Index)
        return i;
    }
    return -1;
  } //enum

  /**
   * Finds ideal spring lengths and spring strengths.
   *
   * @param M The model to work on.
   */
  private void find_l_and_k(Model M)
  {
    int i, j;
    long diam;
    double Lo;
    Node head, tail;
    double headArea, tailArea;

    Ko = 0;
    L = new double[M.getFreeNodeCount()][M.getFreeNodeCount()];
    K = new double[M.getFreeNodeCount()][M.getFreeNodeCount()];
    /* Find the diameter of the graph */
    diam = D[0][0];
    for (i = 0; i < M.getFreeNodeCount(); i++)
    {
      for (j = i + 1; j < M.getFreeNodeCount(); j++)
      {
        if ((diam < D[i][j]) && (D[i][j] < Long.MAX_VALUE))
          diam = D[i][j];
        Ko+= D[i][j];
      }
    }
    Ko = Ko / M.getFreeNodeCount();
    for (i = 0; i < M.getFreeNodeCount(); i++)
    {
      for (j = i + 1; j < M.getFreeNodeCount(); j++)
      {
        /* Lo is desired edge length */
        /*    Lo = avgEdgeLength(G);  */         /* this Lo makes it grow */
        /*  Lo = Math.sqrt(findArea(G))/diam;*/  /* this Lo makes it shrink */
        /* Lo = Math.sqrt(rect.width * rect.height / 4) / diam; */
        head = M.getNodeFromIndex(orderedList[i]);
        tail = M.getNodeFromIndex(orderedList[j]);
        headArea = head.getBoundingBox().height * head.getBoundingBox().width;
        tailArea = tail.getBoundingBox().height * tail.getBoundingBox().width;
        Lo = (Math.sqrt(headArea) + Math.sqrt(tailArea));
        L[i][j] = Lo * D[i][j];
        L[j][i] = L[i][j];
        if (D[i][j] < Long.MAX_VALUE)
        {
          /*  K[i][j] = Ko/(D[i][j]*D[i][j]);*/
          K[i][j] = Ko * Ko / (D[i][j] * D[i][j]);
        }
        else
        {
          K[i][j] = 0;
        }
        K[j][i] = K[i][j];
      } //for (j..
    } //for (i..
  }//find_l_and_k

  /**
   * Finds delta value of a node.
   *
   * @param M The model to work on.
   * @param v The node to get the delta value for.
   * @return The delta value.
   */
  private double findDelta2(Model M, Node v)
  {
    findPartials(M, v);
    return(partial_x * partial_x + partial_y * partial_y);
  } //findDelta2

  /**
   * Finds partial values for delta value.
   *
   * @param M The model to work on.
   * @param v The node to get the values for.
   */
  private void findPartials(Model M, Node v)
  {
    Node u;
    int vi, ui;
    double dx, dy, dd;

    vi = enum(M.getIndexFromNode(v));
    partial_x  = 0;
    partial_y  = 0;
    partial_xx = 0;
    partial_xy = 0;
    partial_yx = 0;
    partial_yy = 0;
    for (u = M.firstFreeNode(); u != null; u = M.nextFreeNode(u))
    {
      if (v != u)
      {
        ui = enum(M.getIndexFromNode(u));
        dx = v.getPosition().x - u.getPosition().x;
        dy = v.getPosition().y - u.getPosition().y;
        dd = Math.sqrt(dx * dx + dy * dy);
        partial_x  += K[vi][ui] * (dx - L[vi][ui] * dx / dd);
        partial_y  += K[vi][ui] * (dy - L[vi][ui] * dy / dd);
        partial_xx += K[vi][ui] * (1 - L[vi][ui] * dy * dy / (dd * dd * dd));
        partial_xy += K[vi][ui] * (L[vi][ui] * dx * dy / (dd * dd * dd));
        partial_yx += K[vi][ui] * (L[vi][ui] * dy * dx / (dd * dd * dd));
        partial_yy += K[vi][ui] * (1 - L[vi][ui] * dx * dx / (dd * dd * dd));
      }
    }
  } //findPartials

  /**
   * Calculates (0.5 * del  +  0.5 * times).
   *
   * @param del
   * @param times
   * @return (0.5 * del  +  0.5 * times).
   */
  private double tempFunction(double del, double times)
  {
    return (0.5 * del  +  0.5 * times);
  } //tempFunction

  /**
   * Finds epsilon value.
   *
   * @param M The model to work on.
   * @return The epsilon value.
   */
  private double findE(Model M)
  {
    Node u, v;
    int vi, ui;
    double dx, dy, dd, dif;
    double total = 0;

    for (v = M.firstFreeNode(); v != null; v = M.nextFreeNode(v))
    {
      vi = enum(M.getIndexFromNode(v));
      for (u = M.nextFreeNode(v); u != null; u = M.nextFreeNode(u))
      {
        ui = enum(M.getIndexFromNode(u));
        dx = v.getPosition().x - u.getPosition().x;
        dy = v.getPosition().y - u.getPosition().y;
        dd = Math.sqrt(dx * dx + dy * dy);
        dif = dd - L[vi][ui];
        total += K[vi][ui] * dif * dif;
      }
    }
    return total / 2;
  } //findE

  /**
   * Moves node mv to new position.
   *
   * @param M The model to work on.
   */
  private void moveToNewPosition(Model M)
  {
    double A, B, C, D, E, F, dx, dy;
    double xpos, ypos, xfrac, yfrac;

    // do not move selected nodes
    if (mv.getSelected())
      return;
    xpos = mv.getPosition().x;
    ypos = mv.getPosition().y;
    int mvi = enum(M.getIndexFromNode(mv));
    findPartials(M, mv);
    A = partial_xx;
    B = partial_xy;
    C = -(partial_x);
    D = partial_yx;
    E = partial_yy;
    F = -(partial_y);
    dy = (A * F - C * D) / (A * E - B * D);
    dx = (C * E - B * F) / (A * E - B * D);
    while ( (dy != 0) && (dx != 0) )
    {
      if (xpos + dx < xmin)
        xfrac = (xpos - xmin) / dx;
      else if (xpos + dx > xmax)
        xfrac = (xmax - xpos) / dx;
      else
        xfrac = 1;
      if (ypos + dy < ymin)
        yfrac = (ypos - ymin) / dy;
      else if (ypos + dy > ymax)
        yfrac = (ymax - ypos) / dy;
      else
        yfrac = 1;
      if (xfrac < yfrac)
      {
        xpos += dx * xfrac;
        dx *= (xfrac - 1);
        ypos += dy * xfrac;
        dy *= (1 - xfrac);
      }
      else if (yfrac < xfrac)
      {
        ypos += dy * yfrac;
        dy *= (yfrac - 1);
        xpos += dx * yfrac;
        dx *= (1 - yfrac);
      }
      else
      {
        xpos += dx;
        dx = 0;
        ypos += dy;
        dy = 0;
      }
    }//end while
    if (mv instanceof EDU.utwente.CoCompose.lang.Composite)
      ((Composite) mv).saveOffsets();
    mv.setPosition(xpos, ypos);
    if (mv instanceof EDU.utwente.CoCompose.lang.Composite)
      ((Composite) mv).restoreOffsets();
  } //moveToNewPosition

  /**
   * Moves node mv to new position.
   *
   * @param M The model to work on.
   */
  private void moveToNewPosition1(Model M)
  {
    double A, B, C, D, E, F, dx, dy;
    double xpos, ypos, xfrac, yfrac;

    // do not move selected nodes
    if (mv.getSelected())
      return;
    xpos = mv.getPosition().x;
    ypos = mv.getPosition().y;
    int mvi = enum(M.getIndexFromNode(mv));
    findPartials(M, mv);
    A = partial_xx;
    B = partial_xy;
    C = -(partial_x);
    D = partial_yx;
    E = partial_yy;
    F = -(partial_y);
    dy = (A * F - C * D) / (A * E - B * D);
    dx = (C * E - B * F) / (A * E - B * D);
    if (xpos + dx < xmin)
      xfrac = (xpos - xmin) / dx;
    else if (xpos + dx > xmax)
      xfrac = (xmax - xpos) / dx;
    else
      xfrac = 1;
    if (ypos + dy < ymin)
      yfrac = (ypos - ymin) / dy;
    else if (ypos + dy > ymax)
      yfrac = (ymax - ypos) / dy;
    else
      yfrac = 1;
    if (xfrac < yfrac)
      yfrac = xfrac;
    else if (yfrac < xfrac)
      xfrac = yfrac;
    xpos += dx * xfrac;
    dx = 0;
    ypos += dy * yfrac;
    dy = 0;
    if (mv instanceof EDU.utwente.CoCompose.lang.Composite)
      ((Composite) mv).saveOffsets();
    mv.setPosition(xpos, ypos);
    if (mv instanceof EDU.utwente.CoCompose.lang.Composite)
      ((Composite) mv).restoreOffsets();
  } //moveToNewPosition1

  /**
   * Moves node mv to new position.
   *
   * @param M The model to work on.
   */
  private void moveToNewPosition2(Model M)
  {
    double A, B, C, D, E, F, dx, dy;
    double xpos, ypos, xfrac, yfrac;

    // do not move selected nodes
    if (mv.getSelected())
      return;
    xpos = mv.getPosition().x;
    ypos = mv.getPosition().y;
    int mvi = enum(M.getIndexFromNode(mv));
    findPartials(M, mv);
    A = partial_xx;
    B = partial_xy;
    C = -(partial_x);
    D = partial_yx;
    E = partial_yy;
    F = -(partial_y);
    dy = (A * F - C * D) / (A * E - B * D);
    dx = (C * E - B * F) / (A * E - B * D);
    if (mv instanceof EDU.utwente.CoCompose.lang.Composite)
      ((Composite) mv).saveOffsets();
    mv.setPosition(xpos + dx, ypos + dy);
    if (mv instanceof EDU.utwente.CoCompose.lang.Composite)
      ((Composite) mv).restoreOffsets();
  } //moveToNewPosition2

  /**
   * Checks if any free nodes share the same position, and if so, moves them
   * away.
   *
   * @param M The model to check.
   */
  private void checkPositions(Model M)
  {
    Node v, u;
    int vi, ui;
    double rand;
    boolean OK = false;

    // move free nodes
    while (!OK)
    {
      OK = true;
      for (v = M.firstFreeNode(); v != null; v = M.nextFreeNode(v))
      {
        for (u = M.nextFreeNode(v); u != null; u = M.nextFreeNode(u))
        {
          // do not move selected nodes
          if (u.getSelected())
            continue;
          // apply check
          if ((v.getPosition().x == u.getPosition().x) &&
              (v.getPosition().y == u.getPosition().y))
          {
            if (u instanceof EDU.utwente.CoCompose.lang.Composite)
              ((Composite) u).saveOffsets();
            vi = enum(M.getIndexFromNode(v));
            ui = enum(M.getIndexFromNode(u));
            rand = (1.4 * Math.random() - 1) * L[vi][ui];
            u.setPosition(u.getPosition().x + rand , u.getPosition().y);
            rand = (1.4 * Math.random() - 1) * L[vi][ui];
            u.setPosition(u.getPosition().x , u.getPosition().y + rand);
            OK = false;
            if (u instanceof EDU.utwente.CoCompose.lang.Composite)
              ((Composite) u).restoreOffsets();
          }
        } //for (u..
      } //for (v..
    } //while (!OK)
  } //checkPositions

  //////////////////////////////////////////////////////////////////////////////
  // Protected methods
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Does one algorithm pass.
   *
   * @return True if the display needs to be updated.
   * @throws Exception if anything goes wrong, which happens if the model
   *                   changes.
   */
  protected boolean doPass() throws Exception
  {
    double tempValue, maxValue, maxDel;
    int maxTimes = 0;
    int mvi, i;
    Node v, u;
    boolean update;

    // get viewport size
    rect = display.windowRect();
    xmin = rect.x;  xmax = xmin + rect.width;
    ymin = rect.y;  ymax = ymin + rect.height;
    // if any pair of vertices share the same position, move them away
    checkPositions(model);
    // determine the next vertex to move according to TempFunction
    maxDel = 0;
    for (v = model.firstFreeNode(); v != null; v = model.nextFreeNode(v))
    {
      delta[enum(model.getIndexFromNode(v))] = Math.sqrt(findDelta2(model, v));
      if (delta[enum(model.getIndexFromNode(v))] > maxDel)
        maxDel = delta[enum(model.getIndexFromNode(v))];
      if (NUM_TIMES_MOVED[enum(model.getIndexFromNode(v))] > maxTimes)
        maxTimes = NUM_TIMES_MOVED[enum(model.getIndexFromNode(v))];
    }
    mv = model.firstFreeNode();
    mvi = enum(model.getIndexFromNode(mv));
    maxValue = 0;
    for (v = mv; v != null; v = model.nextFreeNode(v))
    {
      if (maxTimes != 0)
        tempValue = tempFunction(delta[enum(model.getIndexFromNode(v))] / maxDel,
          1 - NUM_TIMES_MOVED[enum(model.getIndexFromNode(v))] / maxTimes);
      else
        // the first time, base the decision only upon the delta
        tempValue = delta[enum(model.getIndexFromNode(v))] / maxDel;
      if (tempValue > maxValue)
      {
        mv = v;
        mvi = enum(model.getIndexFromNode(mv));
        maxValue = tempValue;
      }
    } //for
    // Check if we have to stop (cancel method invoked).
    if (!running)
      return true;
    // Check if redrawing needed
    if ((COUNTER > HY_SIZE) && (E_HY[COUNTER % HY_SIZE] == (E = findE(model))))
      update = false; // nothing has moved since HY_SIZE iterations already
    else
      update = true;
    // put the E in the history array and get ready to
    // find a new position for vertex 'mv'
    E_HY[COUNTER % HY_SIZE] = E;
    COUNTER++;
    numTimesRepositioned = 0;
    while ( (delta[mvi] > epsilon) &&
            (numTimesRepositioned < MAX_TIMES_REPOSITIONED) )
    {
      //moveToNewPosition(model);  // this function simulates elastic collision
      //  instead of hitting & sticking
      //moveToNewPosition1(model); // this function "hits and sticks"
      moveToNewPosition2(model); // this function doesn't care
      delta[mvi] = Math.sqrt(findDelta2(model, mv));
      numTimesRepositioned++;
    }//end "inner" while
    NUM_TIMES_MOVED[mvi]++;
    return update;
  } //doPass

  //////////////////////////////////////////////////////////////////////////////
  // Public methods
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Do not use. Use the start and stop method instead.
   *
   * @param G The graph to apply this algorithm to.
   * @param update The updatable graph display.
   * @return An error message or null.
   */
  public String compute(Graph G, GraphUpdate update)
  {
    return "This algorithm should be runned using the start() method";
  } //end main compute function

  /**
   * Starts the ModelSpring algorithm and runs it continuously in the background
   * until the stop method is called.
   *
   * @param G The graph to apply this algorithm to.
   * @param update The updatable graph display.
   * @throws CoComposeException if there was an error applying the algorithm.
   */
  public void start(Graph G, GraphUpdate update) throws CoComposeException
  {
    // If the graph isn't a Model, then this algorithm is useless.
    if (!(G instanceof EDU.utwente.CoCompose.lang.Model))
      throw new CoComposeException("This algorithm can only be applied to a Model");
    model = (Model) G;
    display = update;
    // init complete, create and start thread
    running = true;
    thread = new Thread(this);
    thread.setPriority(Thread.MIN_PRIORITY);
    thread.start();
  } //start

  /**
   * Stops the ModelSpring algorithm.
   */
  public void stop()
  {
    // stop run
    running = false;
  } //start

  /**
   * Runs the algorithm. This method should not be called directly. Call start
   * instead.
   */
  public void run()
  {
    int nodeAmount = model.numberOfNodes();
    int edgeAmount = model.numberOfEdges();
    boolean reinit = true;

    while (true)
    {
      try
      {
        // check if we have to do a re-init
        if (reinit)
        {
          nodeAmount = model.numberOfNodes();
          edgeAmount = model.numberOfEdges();
          init(model);
          reinit = false;
        }
        // Do algorithm pass and check if update needed
        if (doPass())
          display.update(false);
        else
          Thread.sleep(500);
      }
      catch (Exception e)
      { // exceptions mean the model has changed, so re-init
        reinit = true;
      }
      // if number of nodes or links in the model changed, re-init
      if ((model.numberOfNodes() != nodeAmount) ||
          (model.numberOfEdges() != edgeAmount))
      {
        reinit = true;
      }
      // Check if we have to stop (cancel method invoked).
      if (!running)
        break;
    }//end "outer" while
  } //run
} //class ModelSpring
