/**
 * Title:        CoCompose
 * Copyright:    Copyright (c) 2002, 2003
 * Author:       Dennis Wagelaar
 * Organisation: University of Twente, Vrije Universiteit Brussel
 * Description:  CoCompose Design Language Tool
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License, version 2 as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  (See the file "COPYING" that is included with this source distribution.)
 */

package EDU.utwente.CoCompose.gui;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.event.*;
import EDU.auburn.VGJ.util.DPoint;

/**
 * A class that allows control of a 3D viewpoint angle in polar coordinates
 * (phi, theta). A StateChanged event is fired when the user drags the crosshair
 * with the mouse.
 * Based upon EDU.auburn.VGJ.gui.AngleControl. Adapted for use with Swing.
 *
 * @author Dennis Wagelaar
 * @author Larry Barowski
 * @version 0.1.10, 22 July 2003
 * @see EDU.auburn.VGJ.gui.AngleControl
 */

public class AngleControl extends JPanel
{
  //////////////////////////////////////////////////////////////////////////////
  // Attributes
  //////////////////////////////////////////////////////////////////////////////

  private int width_ = -1, height_ = -1;
  private double theta_, phi_;
  private double markx_, marky_;
  private Image backImage_;
  private Color backColor_ = Color.white;
  private Color lineColor_ = Color.black;
  private boolean isDragging_;
  /**
   * Only one <code>ChangeEvent</code> is needed per instance since the
   * event's only state is the source property.  The source of events
   * generated is always "this".
   */
  protected transient ChangeEvent changeEvent;

  //////////////////////////////////////////////////////////////////////////////
  // Constructors
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Constructs an AngleControl.
   */
  public AngleControl()
  {
    super();
    try
    {
      jbInit();
      mInit();
    }
    catch (Exception e)
    {
      e.printStackTrace();
    }
  }

  //////////////////////////////////////////////////////////////////////////////
  // Private methods
  //////////////////////////////////////////////////////////////////////////////

  /* JBuilder Component initialization. */
  private void jbInit() throws Exception
  {
    this.setPreferredSize(new Dimension(100, 80));
    this.addMouseMotionListener(new java.awt.event.MouseMotionAdapter()
    {
      public void mouseDragged(MouseEvent e)
      {
        this_mouseDragged(e);
      }
    });
    this.addMouseListener(new java.awt.event.MouseAdapter()
    {
      public void mousePressed(MouseEvent e)
      {
        this_mousePressed(e);
      }
      public void mouseReleased(MouseEvent e)
      {
        this_mouseReleased(e);
      }
    });
  } //jbInit

  /* Manual initialization. */
  private void mInit()
  {
    theta_ = 0.0;
    phi_ = Math.PI / 2.0;
  } //mInit

  /* Re-computes boundaries. */
  private void recompute()
  {
    markx_ = (theta_ + Math.PI) / (2.0 * Math.PI) * (double) width_;
    marky_ = (Math.PI / 2.0 - phi_) / Math.PI * (double) height_;
  } //recompute

  /* Mouse button pressed. */
  void this_mousePressed(MouseEvent e)
  {
    int x = e.getX();
    int y = e.getY();
    // set dragging flag
    isDragging_ = true;
    // process the coordinate move
    processMove(x, y);
  } //this_mousePressed

  /* Mouse button released. */
  void this_mouseReleased(MouseEvent e)
  {
    // set dragging flag
    isDragging_ = false;
    // paint buffer
    Graphics g = getGraphics();
    paintComponentBuffer(g);
    g.dispose();
    // Post an event indicating we're done scrolling.
    fireStateChanged();
  } //this_mouseReleased

  /* Mouse dragged. */
  void this_mouseDragged(MouseEvent e)
  {
    int x = e.getX();
    int y = e.getY();
    // process the coordinate move
    processMove(x, y);
  } //this_mouseDragged

  //////////////////////////////////////////////////////////////////////////////
  // Protected methods
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Paints the angle control elements on top of the panel.
   *
   * @param g The graphics context.
   */
  protected void paintComponent(Graphics g)
  {
    // if resized, discard backbuffer
    Dimension winsize = getSize();
    if (winsize.width != width_ || winsize.height != height_)
    {
      width_ = winsize.width;
      height_ = winsize.height;
      recompute();
      backImage_ = null;
    }
    // create backbuffer if needed
    if (backImage_ == null)
    {
      backImage_ = createImage(width_, height_);
    }
    // get backbuffer graphics context
    Graphics back = backImage_.getGraphics();
    // draw backbuffer
    try
    {
      back.setColor(getBackColor());
      back.fillRect(0, 0, width_, height_);
      back.setColor(getLineColor());
      back.setPaintMode();
      back.drawRect(0, 0, width_ - 1, height_ - 1);
      back.drawLine(width_ / 2, 0, width_ / 2, height_);
      back.drawLine(0, height_ / 2, width_, height_ / 2);
      drawLabels(back, false);
    }
    finally
    {
      back.dispose();
    }
    // draw screen
    paintComponentBuffer(g);
  } //paintComponent

  /**
   * Paints the angle control elements on top of the panel. Does a partial paint
   * by using the backbuffer.
   *
   * @param g The graphics context.
   */
  protected synchronized void paintComponentBuffer(Graphics g)
  {
    Graphics graphics = getComponentGraphics(g);
    try
    {
      graphics.drawImage(backImage_, 0, 0, null);
      graphics.setColor(getLineColor());
      drawX(graphics);
      drawLabels(graphics, true);
    }
    finally
    {
      graphics.dispose();
    }
  }
  /**
   * Draws coordinate system.
   *
   * @param g The graphics context
   */
  protected void drawX(Graphics g)
  {
    g.drawLine((int) markx_ - 3, (int) marky_ - 3,
      (int) markx_ + 3, (int) marky_ + 3);
    g.drawLine((int) markx_ - 3, (int) marky_ + 3,
      (int) markx_ + 3, (int) marky_ - 3);
  } //drawX

  /**
   * Draws the labels.
   *
   * @param g The graphics context
   * @param numbers When true, numbers will be drawn.
   */
  protected void drawLabels(Graphics g, boolean numbers)
  {
    FontMetrics fm = g.getFontMetrics();
    // draw theta
    int thetadeg = (int) (theta_ * 180 / Math.PI);
    String thetastring = "theta";
    if (numbers)
      thetastring += " " + thetadeg;
    g.drawString(thetastring, 2, (height_ / 2 - 1));
    // draw phi
    int phideg = (int) (phi_ * 180 / Math.PI);
    String phistring = "phi";
    if (numbers)
      phistring += " " + phideg;
    g.drawString(phistring, (width_ / 2 + 1), fm.getAscent() + 2);
  } //drawLabels

  /**
   * Notifies all listeners that have registered interest for
   * notification on this event type.  The event instance
   * is lazily created using the parameters passed into
   * the fire method.
   *
   * @see EventListenerList
   */
  protected void fireStateChanged()
  {
    // Guaranteed to return a non-null array
    Object[] listeners = listenerList.getListenerList();
    // Process the listeners last to first, notifying
    // those that are interested in this event
    for (int i = listeners.length - 2; i >= 0; i -= 2)
    {
      if (listeners[i] == ChangeListener.class)
      {
        // Lazily create the event:
        if (changeEvent == null)
          changeEvent = new ChangeEvent(this);
        ((ChangeListener) listeners[i + 1]).stateChanged(changeEvent);
      }
    }
  } //fireStateChanged

  /**
   * Processes a move of mouse coordinates
   *
   * @param x The mouse x coordinate.
   * @param y The mouse y coordinate.
   */
  protected void processMove(int x, int y)
  {
    // clip coordinates
    if (x < 0)
      x = 0;
    else if (x > width_ - 1)
      x = width_ - 1;
    if (y < 0)
      y = 0;
    else if (y > height_ - 1)
      y = height_ - 1;
    // save clipped coordinates
    markx_ = (double) x;
    marky_ = (double) y;
    // set angles
    theta_ = markx_ / (double) (width_ - 1) * 2.0 * Math.PI - Math.PI;
    phi_ = Math.PI / 2.0 - marky_ / (double) (height_ - 1) * Math.PI;
    // paint buffer
    Graphics g = getGraphics();
    paintComponentBuffer(g);
    g.dispose();
    // Post an event indicating a scroll.
    fireStateChanged();
  }

  //////////////////////////////////////////////////////////////////////////////
  // Public methods
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Sets the viewing angles.
   *
   * @param theta The theta angle.
   * @param phi The phi angle.
   */
  public void setAngles(double theta, double phi)
  {
    if (theta == theta_ && phi == phi_)
      return;

    theta_ = theta;
    phi_ = phi;

    recompute();
    repaint();
  } //setAngles

  /**
   * Gets the (theta, phi) viewing angles.
   *
   * @return The (theta, phi) viewing angles.
   */
  public DPoint getAngles()
  {
    return new DPoint(theta_, phi_);
  } //getAngles

  /**
   * Gets the background drawing color.
   *
   * @return The background drawing color.
   */
  public Color getBackColor()
  {
    return backColor_;
  } //getBackColor

  /**
   * Sets the background drawing color.
   *
   * @param color The background drawing color.
   */
  public void setBackColor(Color color)
  {
    backColor_ = color;
  } //setBackColor

  /**
   * Gets the line drawing color.
   *
   * @return The line drawing color.
   */
  public Color getLineColor()
  {
    return lineColor_;
  } //getLineColor

  /**
   * Sets the line drawing color.
   *
   * @param color The line drawing color.
   */
  public void setLineColor(Color color)
  {
    lineColor_ = color;
  } //setLineColor

  /**
   * Returns whether the mouse is dragging.
   *
   * @return True if and only if the mouse is dragging.
   */
  public boolean isDragging()
  {
    return isDragging_;
  } //isDragging

  /**
   * Adds a ChangeListener to the viewport scroller.
   *
   * @param l the listener to be added
   */
  public void addChangeListener(ChangeListener l)
  {
    listenerList.add(ChangeListener.class, l);
  } //addChangeListener

  /**
   * Removes a ChangeListener from the viewport scroller.
   *
   * @param l the listener to be removed
   */
  public void removeChangeListener(ChangeListener l)
  {
    listenerList.remove(ChangeListener.class, l);
  } //removeChangeListener
} //class AngleControl
