/**
 * Title:        CoCompose
 * Copyright:    Copyright (c) 2002
 * Author:       Dennis Wagelaar
 * Organisation: University of Twente, Vrije Universiteit Brussel
 * Description:  CoCompose Design Language Tool
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License, version 2 as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  (See the file "COPYING" that is included with this source distribution.)
 */

package EDU.utwente.CoCompose.gui;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.event.*;
import EDU.auburn.VGJ.util.DDimension;
import EDU.auburn.VGJ.util.DPoint;
import java.beans.*;

/**
 * A panel containing a ModelCanvas and a header label. Based upon
 * EDU.auburn.VGJ.gui.ScrolledPanel. Adapted to support Swing.
 *
 * @author Dennis Wagelaar
 * @author Larry Barowski
 * @version 0.0.17, 3 April 2002
 * @see EDU.auburn.VGJ.gui.ScrolledPanel
 */

public class ModelPanel extends JPanel
{
  //////////////////////////////////////////////////////////////////////////////
  // Constants
  //////////////////////////////////////////////////////////////////////////////

  private static final double PAGE_FRACTION = 0.25;

  //////////////////////////////////////////////////////////////////////////////
  // Attributes
  //////////////////////////////////////////////////////////////////////////////

  private double xoffset_ = 0.0;
  private double yoffset_ = 0.0;
  private DDimension portSize_;
  private DDimension contentSize_;
  private ModelCanvas modelCanvas_;
  private boolean dontFireEvent_;
  /**
   * Only one <code>ChangeEvent</code> is needed per instance since the
   * event's only state is the source property.  The source of events
   * generated is always "this".
   */
  protected transient ChangeEvent changeEvent;
  JPanel labelPanel = new JPanel();
  BorderLayout borderLayout1 = new BorderLayout();
  JLabel label = new JLabel();
  FlowLayout flowLayout1 = new FlowLayout();
  JPanel canvasPanel = new JPanel();
  BorderLayout borderLayout2 = new BorderLayout();
  JScrollBar vScrollBar = new JScrollBar();
  JPanel bottomPanel = new JPanel();
  BorderLayout borderLayout3 = new BorderLayout();
  JScrollBar hScrollBar = new JScrollBar();
  JPanel fillerPanel = new JPanel();

  //////////////////////////////////////////////////////////////////////////////
  // Constructors
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Constructs a ModelPanel.
   *
   * @param modelCanvas The model canvas to embed.
   */
  public ModelPanel(ModelCanvas modelCanvas)
  {
    super();
    modelCanvas_ = modelCanvas;
    try
    {
      jbInit();
      mInit();
    }
    catch(Exception e)
    {
      e.printStackTrace();
    }
  } //ModelPanel

  /**
   * Constructs a ModelPanel.
   * Required by the JBuilder UI designer, do not use!
   */
  public ModelPanel()
  {
    this(new ModelCanvas());
  } //ModelPanel()

  //////////////////////////////////////////////////////////////////////////////
  // Private methods
  //////////////////////////////////////////////////////////////////////////////

  /* JBuilder Component initialization. */
  private void jbInit() throws Exception
  {
    this.setLayout(borderLayout1);
    labelPanel.setLayout(flowLayout1);
    flowLayout1.setAlignment(FlowLayout.LEFT);
    canvasPanel.setLayout(borderLayout2);
    bottomPanel.setLayout(borderLayout3);
    hScrollBar.setOrientation(JScrollBar.HORIZONTAL);
    hScrollBar.addAdjustmentListener(new java.awt.event.AdjustmentListener()
    {
      public void adjustmentValueChanged(AdjustmentEvent e)
      {
        hScrollBar_adjustmentValueChanged(e);
      }
    });
    fillerPanel.setMinimumSize(new Dimension(16, 16));
    fillerPanel.setPreferredSize(new Dimension(16, 16));
    canvasPanel.setBorder(BorderFactory.createLoweredBevelBorder());
    modelCanvas_.addChangeListener(new javax.swing.event.ChangeListener()
    {
      public void stateChanged(ChangeEvent e)
      {
        modelCanvas__stateChanged(e);
      }
    });
    modelCanvas_.addPropertyChangeListener(new java.beans.PropertyChangeListener()
    {
      public void propertyChange(PropertyChangeEvent e)
      {
        modelCanvas__propertyChange(e);
      }
    });
    vScrollBar.addAdjustmentListener(new java.awt.event.AdjustmentListener()
    {
      public void adjustmentValueChanged(AdjustmentEvent e)
      {
        vScrollBar_adjustmentValueChanged(e);
      }
    });
    label.setText(" ");
    this.add(labelPanel,  BorderLayout.NORTH);
    labelPanel.add(label, null);
    this.add(canvasPanel, BorderLayout.CENTER);
    canvasPanel.add(vScrollBar,  BorderLayout.EAST);
    canvasPanel.add(bottomPanel,  BorderLayout.SOUTH);
    bottomPanel.add(hScrollBar,  BorderLayout.CENTER);
    bottomPanel.add(fillerPanel, BorderLayout.EAST);
  } //jbInit

  /* Manual initialization. */
  private void mInit()
  {
    canvasPanel.add(modelCanvas_, BorderLayout.CENTER);
  } //mInit

  /* The state of the ModelCanvas has changed. */
  void modelCanvas__stateChanged(ChangeEvent e)
  {
    configure();
    // post an event
    fireStateChanged();
  } //modelCanvas__stateChanged

  /* A property of the ModelCanvas has changed. */
  void modelCanvas__propertyChange(PropertyChangeEvent e)
  {
    if (e.getPropertyName().equals("Label"))
    {
      String string = (String) e.getNewValue();
      label.setText(string);
    }
  } //modelCanvas__propertyChange

  /* Horizontal scrollbar position changed. */
  void hScrollBar_adjustmentValueChanged(AdjustmentEvent e)
  {
    xoffset_ = (double) hScrollBar.getValue();
    modelCanvas_.setOffset(xoffset_, yoffset_, true);
    // post an event if needed
    if (!dontFireEvent_)
      fireStateChanged();
  } //hScrollBar_adjustmentValueChanged

  /* Vertical scrollbar position changed. */
  void vScrollBar_adjustmentValueChanged(AdjustmentEvent e)
  {
    yoffset_ = (double) vScrollBar.getValue();
    modelCanvas_.setOffset(xoffset_, yoffset_, true);
    // post an event if needed
    if (!dontFireEvent_)
      fireStateChanged();
  } //vScrollBar_adjustmentValueChanged

  //////////////////////////////////////////////////////////////////////////////
  // Protected methods
  //////////////////////////////////////////////////////////////////////////////

   /**
    * Calculate new Scrollbar values, maximums, and sizes, and adjust the
    * offset of the OffsetCanvas if necessary.
    */
   protected void configure()
   {
    double pageh, pagew;
    Dimension tmpD = modelCanvas_.getSize();

    portSize_ = new DDimension(tmpD.width, tmpD.height);
    contentSize_ = modelCanvas_.getContentSize();

    DPoint tmpdim = modelCanvas_.getOffset();
    xoffset_ = tmpdim.x;
    yoffset_ = tmpdim.y;

    // Set scrollbar bounds
    hScrollBar.setValues((int) xoffset_, (int) portSize_.width, 0, (int) contentSize_.width);
    vScrollBar.setValues((int) yoffset_, (int) portSize_.height, 0, (int) contentSize_.height);

    // page increment is pageFraction_ of screen size but at least one
    pageh = Math.max((portSize_.width * PAGE_FRACTION), 1.0);
    pagew = Math.max((portSize_.height * PAGE_FRACTION), 1.0);
    hScrollBar.setBlockIncrement((int) pageh);
    vScrollBar.setBlockIncrement((int) pagew);
  } //configure

  /**
   * Notifies all listeners that have registered interest for
   * notification on this event type.  The event instance
   * is lazily created using the parameters passed into
   * the fire method.
   *
   * @see EventListenerList
   */
  protected void fireStateChanged()
  {
    // Guaranteed to return a non-null array
    Object[] listeners = listenerList.getListenerList();
    // Process the listeners last to first, notifying
    // those that are interested in this event
    for (int i = listeners.length - 2; i >= 0; i -= 2)
    {
      if (listeners[i] == ChangeListener.class)
      {
        // Lazily create the event:
        if (changeEvent == null)
          changeEvent = new ChangeEvent(this);
        ((ChangeListener) listeners[i + 1]).stateChanged(changeEvent);
      }
    }
  } //fireStateChanged

  //////////////////////////////////////////////////////////////////////////////
  // Public methods
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Gets the viewport size.
   *
   * @return The viewport size.
   */
  public DDimension getPortSize()
  {
    return new DDimension(portSize_.width, portSize_.height);
  } //getPortSize

  /**
   * Gets the content size.
   *
   * @return The content size.
   */
  public DDimension getContentSize()
  {
    return new DDimension(contentSize_.width, contentSize_.height);
  } //getContentSize

  /**
   * Gets the scrolling offset.
   *
   * @return The scrolling offset
   */
  public DPoint getOffset()
  {
    return new DPoint(xoffset_, yoffset_);
  } //getOffset

  /**
   * Scroll to the given location.
   *
   * @param x Horizontal offset to scroll to.
   * @param y Vertical offset to scroll to.
   */
  public void scrollTo(double x, double y)
  {
    // clip lower bound
    if (x < 0.0)
      x = 0.0;
    if (y < 0.0)
      y = 0.0;
    // clip upper bound
    if (x > (double) hScrollBar.getMaximum())
      x = (double) hScrollBar.getMaximum();
    if (y > (double) vScrollBar.getMaximum())
      y = (double) vScrollBar.getMaximum();
    // set internal offset attribute
    xoffset_ = x;
    yoffset_ = y;
    // set scrollbars (without triggering events)
    dontFireEvent_ = true;
    hScrollBar.setValue((int) x);
    vScrollBar.setValue((int) y);
    dontFireEvent_ = false;
    // adjust canvas
    modelCanvas_.setOffset(x, y, true);
  } //scrollTo

  /**
   * Adds a ChangeListener to the viewport scroller.
   *
   * @param l the listener to be added
   */
  public void addChangeListener(ChangeListener l)
  {
    listenerList.add(ChangeListener.class, l);
  } //addChangeListener

  /**
   * Removes a ChangeListener from the viewport scroller.
   *
   * @param l the listener to be removed
   */
  public void removeChangeListener(ChangeListener l)
  {
    listenerList.remove(ChangeListener.class, l);
  } //removeChangeListener
} //class ModelPanel
