/**
 * Title:        CoCompose
 * Copyright:    Copyright (c) 2002, 2003
 * Author:       Dennis Wagelaar
 * Organisation: University of Twente, Vrije Universiteit Brussel
 * Description:  CoCompose Design Language Tool
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License, version 2 as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  (See the file "COPYING" that is included with this source distribution.)
 */

package EDU.utwente.CoCompose.lang;

import java.util.*;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.DefaultHandler;
import org.xml.sax.helpers.AttributesImpl;
import EDU.auburn.VGJ.graph.Node;
import EDU.utwente.CoCompose.CoCompose;
import EDU.utwente.CoCompose.translate.Implementation;
import EDU.utwente.CoCompose.util.*;

/**
 * The CoCompose design language primitive Concept.
 *
 * @author Dennis Wagelaar
 * @version 0.1.17, 19 November 2003
 */

public class Concept extends AbstractConcept
{
  //////////////////////////////////////////////////////////////////////////////
  // Constants
  //////////////////////////////////////////////////////////////////////////////

  private static final String XMLNAME_IMPLEMENTATIONS = "Implementations";

  //////////////////////////////////////////////////////////////////////////////
  // Attributes
  //////////////////////////////////////////////////////////////////////////////

  // The set of implementation for this Concept.
  private Vector implementations_ = new Vector();
  // The concept form, if determined
  private String form_ = new String();
  // The implementation for the concept form, if determined
  private Implementation formImpl_ = null;

  //////////////////////////////////////////////////////////////////////////////
  // Constructors
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Constructs a Concept.
   */
  public Concept()
  {
    super();
  } //Concept

  //////////////////////////////////////////////////////////////////////////////
  // Protected methods
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Create XML from the implementations using a SAX DefaultHandler.
   *
   * @param handler The SAX DefaultHandler.
   * @throws SAXException if an XML handler error occurs.
   */
  protected void implementationsToXML(DefaultHandler handler) throws SAXException
  {
    Enumeration implementations = getImplementations();
    // if no implementations, return
    if (!implementations.hasMoreElements())
      return;

    handler.startElement(null, null, XMLNAME_IMPLEMENTATIONS, null);
    while (implementations.hasMoreElements())
    {
      ((Implementation) implementations.nextElement()).toXML(handler);
    }
    handler.endElement(null, null, XMLNAME_IMPLEMENTATIONS);
  } //implementationsToXML

  /**
   * Create XML start from this object using a SAX DefaultHandler, element name
   * and given attributes.
   *
   * @param handler The SAX DefaultHandler.
   * @param element The XML element name.
   * @param atts The XML element attributes.
   * @throws SAXException if an XML handler error occurs.
   */
  protected void startXML(DefaultHandler handler, String element,
    AttributesImpl atts) throws SAXException
  {
    // create new attributes list if none given.
    if (atts == null)
      atts = new AttributesImpl();

    // Add attributes
    UniqueRealm realm = getUniqueRealm();
    Model model = getModel();
    if (realm != null && realm != model)
    {
      atts.addAttribute(null, null, "From", "CDATA", realm.getName());
    }
    String form = getForm();
    if (form.length() > 0)
    {
      atts.addAttribute(null, null, "Form", "CDATA", form);
    }

    // Write XML
    super.startXML(handler, element, atts);

    // Implementations
    implementationsToXML(handler);
  } //startXML

  //////////////////////////////////////////////////////////////////////////////
  // Public methods
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Sets the label using the object attributes.
   */
  public synchronized void setLabel()
  {
    UniqueRealm realm = getUniqueRealm();
    Model model = getModel();
    String form = getForm();
    String namespace = new String();
    // display namespace if not current model
    if (realm != null && realm != model)
    {
      namespace = "\\n(from " + realm.getName();
      if (namespace.length() > 16)
        namespace = namespace.substring(0, 16) + "...";
      namespace += ")";
    }
    // display concept form if not null
    if (form.length() > 0)
    {
      form = "\\n(as " + form + ")";
    }
    // set label
    super.setLabel();
    setLabel(getLabel() + namespace + form);
  } //setLabel

  /**
   * Adds an implementation for this concept.
   *
   * @param implementation The implementation to add.
   */
  public void addImplementation(Implementation implementation)
  {
    implementations_.add(implementation);
    changeListeners.add(implementation);
    implementation.setConcept(this);
  } //addImplementation

  /**
   * Inserts an implementation for this concept on a given index.
   *
   * @param implementation The implementation to insert.
   * @param index The index on which to insert the implementation.
   */
  public void insertImplementation(Implementation implementation, int index)
  {
    implementations_.insertElementAt(implementation, index);
    implementation.setConcept(this);
  } //insertImplementation

  /**
   * Removes an implementation for this concept.
   *
   * @param implementation The implementation to remove.
   */
  public void removeImplementation(Implementation implementation)
  {
    implementation.setConcept(null);
    changeListeners.remove(implementation);
    implementations_.remove(implementation);
  } //removeImplementation

  /**
   * Gets the implementations for this concept.
   *
   * @return The implementations for this concept.
   */
  public Enumeration getImplementations()
  {
    return implementations_.elements();
  } //getImplementations()

  /**
   * Gets the implementations for this concept given the language.
   *
   * @param language The language in which the implementations should be
   *                 expressed.
   * @return The implementations in the given language for this concept.
   */
  public Enumeration getImplementations(String language)
  {
    Implementation implementation;
    Vector langImplementations = new Vector();

    Enumeration implementations = implementations_.elements();
    while (implementations.hasMoreElements())
    {
      implementation = (Implementation) implementations.nextElement();
      String thisLanguage = language;
      while (thisLanguage != null)
      {
        if (implementation.getLanguage().equals(thisLanguage))
          langImplementations.add(implementation);
        thisLanguage = CoCompose.getSuperLanguage(thisLanguage);
      }
    }
    return langImplementations.elements();
  } //getImplementations(String language)

  /**
   * Gets the amount of implementations this concept has.
   *
   * @return The amount of implementations.
   */
  public int getImplementationCount()
  {
    return implementations_.size();
  } //getImplementationCount

  /**
   * Gets the concept form, if determined.
   *
   * @return The concept form, if determined.
   */
  public String getForm()
  {
    return form_;
  } //getForm

  /**
   * Sets the concept form, if determined.
   *
   * @param form The concept form.
   */
  public void setForm(String form)
  {
    form_ = form;
    setLabel();
  } //setForm

  /**
   * Copies attributes from fromConcept to self. Inherited Node attributes are
   * NOT copied.
   *
   * @param fromConcept The Concept to copy the attributes from.
   */
  public void copyAttributes(Concept fromConcept)
  {
    Implementation implementationCopy, thisImplementation;

    // set inherited attributes
    super.copyAttributes(fromConcept);
    // set own attributes
    Enumeration implementations = fromConcept.getImplementations();
    while (implementations.hasMoreElements())
    {
      thisImplementation = (Implementation) implementations.nextElement();
      implementationCopy = new Implementation();
      implementationCopy.copyAttributes(thisImplementation);
      addImplementation(implementationCopy);
    }
    setLabel();
  } //copyAttributes

  /**
   * Copies implementation constraints from fromConcept to self.
   * @param fromConcept the Concept to copy the constraints from.
   * @throws CoComposeException - if one or more constraints could not be copied.
   */
  public void copyConstraints(Concept fromConcept)
    throws CoComposeException
  {
    Implementation impl, fromImpl;

    // copy constraints per implementation
    Enumeration fromImpls = fromConcept.getImplementations();
    Enumeration impls = getImplementations();
    while (impls.hasMoreElements() && fromImpls.hasMoreElements())
    {
      impl = (Implementation) impls.nextElement();
      fromImpl = (Implementation) fromImpls.nextElement();
      impl.copyConstraints(fromImpl);
    }
  } //copyConstraints

  /**
   * Copies implementation constraints from fromConcept to self using the
   * given unique realms in which to find the constraints.
   *
   * @param fromConcept The Concept to copy the constraints from.
   * @param inRealms The unique realms in which to find the constraints.
   * @throws CoComposeException if one or more constraints could not be copied.
   */
/*  public void copyConstraints(Concept fromConcept, Set inRealms)
    throws CoComposeException
  {
    Implementation impl, fromImpl;

    // copy constraints per implementation
    Enumeration fromImpls = fromConcept.getImplementations();
    Enumeration impls = getImplementations();
    while (impls.hasMoreElements() && fromImpls.hasMoreElements())
    {
      impl = (Implementation) impls.nextElement();
      fromImpl = (Implementation) fromImpls.nextElement();
      impl.copyConstraints(fromImpl, inRealms);
    }
  } //copyConstraints */

  /**
   * Retrieves all Concepts linked to this Concept through a composite.
   *
   * @param logger The object to log to.
   * @return The collection of neighbour concepts.
   * @throws Exception if neighbours could not be retrieved.
   */
  public Set getNeighbourConcepts(Logger logger) throws Exception
  {
    HashSet neighbours = new HashSet();
    Model model = getModel();
    Node node, subNode;
    Composite composite;
    Collection linkedNodes;
    Iterator links;

    logger.log("Retrieving neighbour concepts for " + getName());
    logger.incLogLevel();
    try
    {
      // check model presence
      if (model == null)
      {
        logger.log("Error: model not set for " + getName());
        throw new Exception("Model not set for " + getName());
      }
      // retrieve linked composites
      for (int child = firstChild(); child >= 0; child = nextChild())
      {
        node = model.getNodeFromIndex(child);
        // check for composite or composite role
        composite = null;
        if (node instanceof EDU.utwente.CoCompose.lang.Composite)
          composite = (Composite) node;
        else if (node instanceof EDU.utwente.CoCompose.lang.CompositeRole)
          composite = ((CompositeRole) node).getComposite();
        // if a composite link is retrieved, retrieve all concepts linked to that
        // composite
        if (composite != null)
        {
          linkedNodes = model.getLinkedNodes(composite);
          // retrieve all Concepts from these
          links = linkedNodes.iterator();
          while (links.hasNext())
          {
            subNode = (Node) links.next();
            if ((subNode instanceof EDU.utwente.CoCompose.lang.Concept) &&
                (!subNode.equals(this)))
            {
              neighbours.add((Concept) subNode);
              logger.log("Neighbour " + ((Concept) subNode).getName() + " added");
            }
          } //while links
        } //if composite
      } //for child
    }
    finally
    {
      logger.decLogLevel();
    }
    return neighbours;
  } //getNeighbourConcepts

  /**
   * Accepts a ModelVisitor.
   *
   * @param visitor The visitor object to accept.
   * @throws CoComposeException if visiting operation failed.
   */
  public void accept(ModelVisitor visitor) throws CoComposeException
  {
    super.accept(visitor);
    visitor.visitConcept(this);
  } //accept
} //class Concept
