/**
 * Title:        CoCompose
 * Copyright:    Copyright (c) 2002, 2003
 * Author:       Dennis Wagelaar
 * Organisation: University of Twente, Vrije Universiteit Brussel
 * Description:  CoCompose Design Language Tool
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License, version 2 as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  (See the file "COPYING" that is included with this source distribution.)
 */

package EDU.utwente.CoCompose.lang;

import java.beans.PropertyChangeEvent;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.DefaultHandler;
import org.xml.sax.helpers.AttributesImpl;
import EDU.auburn.VGJ.util.*;
import EDU.utwente.CoCompose.util.*;

/**
 * The role class representing the roles of the CoCompose design language
 * primitive Composite.
 *
 * @author Dennis Wagelaar
 * @version 0.1.10, 22 July 2003
 */

public class PublishedConcept extends CompositePartConcept
{
  //////////////////////////////////////////////////////////////////////////////
  // Constants
  //////////////////////////////////////////////////////////////////////////////

  private static final String XMLNAME_PUBLISHED = "PublishedConcept";

  //////////////////////////////////////////////////////////////////////////////
  // Attributes
  //////////////////////////////////////////////////////////////////////////////

  private Identifier publicName_ = null;

  //////////////////////////////////////////////////////////////////////////////
  // Constructors
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Constructs a PublishedConcept.
   */
  public PublishedConcept()
  {
    super();
  } //PublishedConcept

  //////////////////////////////////////////////////////////////////////////////
  // Protected methods
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Bounds one dimension of a position only.
   *
   * @param new_pos A new position coordinate dimension.
   * @param size The corresponding dimension size.
   * @param bound_pos The boundary position coordinate dimension.
   * @param bound_size The corresponding dimension boundary size.
   * @return The clipped position coordinate dimension.
   */
  protected double boundDimension(double new_pos, double size,
    double bound_pos, double bound_size)
  {
    if (new_pos < bound_pos - bound_size / 2)
    {
      new_pos = bound_pos - bound_size / 2;
    }
    else if (new_pos > bound_pos + bound_size / 2)
    {
      new_pos = bound_pos + bound_size / 2;
    }
    return new_pos;
  } //boundDimension

  /**
   * Bounds a new role position within the bounds of the composite.
   *
   * @param new_position The new position coordinate to clip.
   * @return The clipped position coordinate.
   */
  protected DPoint boundPosition(DPoint new_position)
  {
    Composite composite = getComposite();
    DPoint bound_position = new DPoint();
    double dx, dy;

    if (composite != null)
    {
      // get composite position and size
      DDimension dim = getBoundingBox();
      DPoint compositePos = composite.getPosition();
      DDimension compositeDim = composite.getBoundingBox();
      // check x coordinate bounds
      bound_position.x = boundDimension(new_position.x, dim.width, compositePos.x,
        compositeDim.width);
      // check y coordinate bounds
      bound_position.y = boundDimension(new_position.y, dim.height, compositePos.y,
        compositeDim.height);
      // either x or y must lie on the composite's bound
      if ((bound_position.x == new_position.x) && (bound_position.y == new_position.y))
      { // x and y do not lie on bound
        dx = Math.abs(compositePos.x - bound_position.x) / compositeDim.width;
        dy = Math.abs(compositePos.y - bound_position.y) / compositeDim.height;
        // highest distance to center (relative to size) clips
        if (dx > dy)
        { // clip x
          if (bound_position.x > compositePos.x)
            bound_position.x = compositePos.x + compositeDim.width / 2;
          else
            bound_position.x = compositePos.x - compositeDim.width / 2;
        }
        else
        { // clip y
          if (bound_position.y > compositePos.y)
            bound_position.y = compositePos.y + compositeDim.height / 2;
          else
            bound_position.y = compositePos.y - compositeDim.height / 2;
        }
      } //if bound_position
    }
    return bound_position;
  } //boundPosition(DPoint new_position)

  /**
   * Bounds a new role position within the bounds of the composite.
   *
   * @param new_position The new position coordinate to clip.
   * @return The clipped position coordinate.
   */
  protected DPoint3 boundPosition(DPoint3 new_position)
  {
    Composite composite = getComposite();
    DPoint3 bound_position = new DPoint3();
    double dx, dy;

    if (composite != null)
    {
      // get composite position and size
      DDimension3 dim = getBoundingBox3();
      DPoint3 compositePos = composite.getPosition3();
      DDimension3 compositeDim = composite.getBoundingBox3();
      // check x coordinate bounds
      bound_position.x = boundDimension(new_position.x, dim.width, compositePos.x,
        compositeDim.width);
      // check y coordinate bounds
      bound_position.y = boundDimension(new_position.y, dim.height, compositePos.y,
        compositeDim.height);
      // check z coordinate bounds
      bound_position.z = boundDimension(new_position.z, dim.depth, compositePos.z,
        compositeDim.depth);
      // either x or y must lie on the composite's bound (z may lie anywhere)
      if ((bound_position.x == new_position.x) && (bound_position.y == new_position.y))
      { // x and y do not lie on bound
        dx = Math.abs(compositePos.x - bound_position.x) / compositeDim.width;
        dy = Math.abs(compositePos.y - bound_position.y) / compositeDim.height;
        // highest distance to center (relative to size) clips
        if (dx > dy)
        { // clip x
          if (bound_position.x > compositePos.x)
            bound_position.x = compositePos.x + compositeDim.width / 2;
          else
            bound_position.x = compositePos.x - compositeDim.width / 2;
        }
        else
        { // clip y
          if (bound_position.y > compositePos.y)
            bound_position.y = compositePos.y + compositeDim.height / 2;
          else
            bound_position.y = compositePos.y - compositeDim.height / 2;
        }
      } //if bound_position
    }
    return bound_position;
  } //boundPosition(DPoint3 new_position)

  /**
   * Create XML start from this object using a SAX DefaultHandler, element name
   * and given attributes.
   *
   * @param handler The SAX DefaultHandler.
   * @param element The XML element name.
   * @param atts The XML element attributes.
   * @throws SAXException if an XML handler error occurs.
   */
  protected void startXML(DefaultHandler handler, String element,
    AttributesImpl atts) throws SAXException
  {
    // The method needs to be rewritten from AbstractConcept, since
    // PartOf attribute needs to be suppressed.
    // create new attributes list if none given.
    if (atts == null)
      atts = new AttributesImpl();

    // Add attributes
    atts.addAttribute(null, null, "CompletenessConstraint", "CDATA",
      getCompletenessConstraint());
    atts.addAttribute(null, null, "CorrectnessConstraint", "CDATA",
      getCorrectnessConstraint());
    atts.addAttribute(null, null, "Id", "ID", String.valueOf(getId()));
    atts.addAttribute(null, null, "Name", "CDATA", getName());
    if (getPublicName().length() > 0)
      atts.addAttribute(null, null, "PublicName", "CDATA", getPublicName());

    // Write XML
    handler.startElement(null, null, element, atts);

    // GML
    gmlToXML(handler);

    // Comments
    commentsToXML(handler);

    // Specification
    specToXML(handler);
  } //startXML

  /**
   * Notfies the ProprtyChangeListeners of a name change. Stub in this class,
   * because the normal name is not public within the model.
   * @param oldName The old name property value.
   * @param newName The new name property value.
   */
  protected void notifyNameChange(String oldName, String newName)
  {
  } //notifyNameChange

  /**
   * Notfies the ProprtyChangeListeners of a public name change.
   * @param oldName The old publicName property value.
   * @param newName The new publicName property value.
   */
  protected void notifyPublicNameChange(String oldName, String newName)
  {
    notifyListeners(new PropertyChangeEvent(this, "Name", oldName, newName));
  } //notifyPublicNameChange

  //////////////////////////////////////////////////////////////////////////////
  // Public methods
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Sets the defaults for this published concept.
   */
  public void setDefaults()
  {
    defaultLabel_ = "Published";
    setShape(DefaultShape);
    setLabelPosition(DefaultLabelPos);
    // set default colors
    setEdgeColor(DEFAULT_EDGE_COLOR);
    setFillColor(DEFAULT_FILL_COLOR);
    setFontColor(DEFAULT_FONT_COLOR);
    setSelectColor(DEFAULT_SELECT_COLOR);
  } //setDefaults

  /**
   * Gets the unique realm namespace for the public name. This is usually the
   * unique realm in which the composite resides.
   *
   * @return The unique realm or null.
   */
  public UniqueRealm getPublicUniqueRealm()
  {
    Composite composite = getComposite();
    UniqueRealm realm = null;

    if (composite != null)
    { // composite is set, so we can use its realm to enforce uniqueness in the
      // model.
      realm = composite.getUniqueRealm();
    }
    return realm;
  } //getPublicUniqueRealm

  /**
   * Gets the public name attribute.
   *
   * @return The public name attribute.
   */
  public String getPublicName()
  {
    if (publicName_ != null)
      return publicName_.getValue();
    else
      return "";
  } //getPublicName

  /**
   * Sets the public name attribute - sets label as well.
   *
   * @param name_in The new public name.
   * @throws CoComposeException if the name is not a valid unique identifier or
   *                            if the Composite is not set for this.
   */
  public synchronized void setPublicName(String name_in) throws CoComposeException
  {
    UniqueRealm realm = getPublicUniqueRealm();

    if (publicName_ == null)
    {
      notifyPublicNameChange("", name_in);
      if (name_in.length() > 0)
        publicName_ = new Identifier(realm, name_in);
    }
    else
    {
      notifyPublicNameChange(publicName_.getValue(), name_in);
      if (name_in.length() == 0)
      {
        if (realm != null)
          realm.unRegister(publicName_.getValue());
        publicName_ = null;
      }
      else
      {
        publicName_.setValue(name_in);
      }
    }
    setLabel();
  } //setPublicName

  /**
   * Sets the label using the object attributes.
   */
  public synchronized void setLabel()
  {
    // set label depending on public name.
    if (getPublicName().length() > 0)
    {
      setLabel(getPublicName());
    }
    else
    {
      super.setLabel();
    }
  } //setLabel

  /**
   * Create XML from this object using a SAX DefaultHandler.
   *
   * @param handler The SAX DefaultHandler.
   * @throws SAXException if an XML handler error occurs.
   */
  public void toXML(DefaultHandler handler) throws SAXException
  {
    // Write XML
    startXML(handler, XMLNAME_PUBLISHED, null);
    endXML(handler, XMLNAME_PUBLISHED);
  } //toXML

  /**
   * Accepts a ModelVisitor.
   *
   * @param visitor The visitor object to accept.
   * @throws CoComposeException if visiting operation failed.
   */
  public void accept(ModelVisitor visitor) throws CoComposeException
  {
    super.accept(visitor);
    visitor.visitPublishedConcept(this);
  } //accept
} //class PublishedConcept
