/**
 * Title:        CoCompose
 * Copyright:    Copyright (c) 2002, 2003
 * Author:       Dennis Wagelaar
 * Organisation: University of Twente, Vrije Universiteit Brussel
 * Description:  CoCompose Design Language Tool
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License, version 2 as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  (See the file "COPYING" that is included with this source distribution.)
 */

package EDU.utwente.CoCompose.translate;

import java.io.*;
import java.util.*;
import EDU.utwente.CoCompose.lang.*;
import EDU.utwente.CoCompose.util.*;

/**
 * Abstract Generate Implementation visitor for generating implementation from
 * CoCompose Models.
 *
 * @author Dennis Wagelaar
 * @version 0.1.15, 20 October 2003
 */

public abstract class GenerateImplementationVisitor extends ModelVisitor
{
  //////////////////////////////////////////////////////////////////////////////
  // Constants
  //////////////////////////////////////////////////////////////////////////////

  // Java package path to plugin implementation generator classes
  private static final String IMPLGEN_PATH = "implementations.";
  /**
   * Indentation string (two spaces by default).
   */
  public static final String Indent = "  ";
  /**
   * File output directory for all generated implementation files.
   */
  public static final String OUTPUT_DIR = "./output/";

  //////////////////////////////////////////////////////////////////////////////
  // Attributes
  //////////////////////////////////////////////////////////////////////////////

  // The concept forms in the model.
  // Structure: Concept |-> String
  private Hashtable forms_ = new Hashtable();
  // Concept/UniqueRealm to implementation mapping.
  // Structure: (Concept | UniqueRealm) |-> String
  private Hashtable impls_ = new Hashtable();
  /**
   * The form checking visitor needed to check constraints
   */
  protected FormConstraintsVisitor constraintCheck = null;

  //////////////////////////////////////////////////////////////////////////////
  // Constructors
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Constructs a GenerateImplementationVisitor.
   */
  public GenerateImplementationVisitor()
  {
  } //GenerateImplementationVisitor

  //////////////////////////////////////////////////////////////////////////////
  // Protected methods
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Gets the concept forms in the model: Concept |-> String.
   *
   * @return The concept forms per concept.
   */
  protected Hashtable getForms()
  {
    return forms_;
  } //getForms

  /**
   * Sets a concept form for a concept.
   *
   * @param concept The concept to set the form for.
   * @param form The concept form to set.
   */
  protected void setForm(Concept concept, String form)
  {
    forms_.put(concept, form);
  } //setForm

  /**
   * Removes a concept form mapping.
   *
   * @param concept The concept to remove.
   */
  protected void removeForm(Concept concept)
  {
    forms_.remove(concept);
  } //removeForm

  /**
   * Gets the implementations for the concepts and unique realms.
   *
   * @return The implementations.
   * Structure: (Concept | UniqueRealm) |-> String.
   */
  protected Hashtable getImpls()
  {
    return impls_;
  } //getImpls

  /**
   * Gets the first working implementation for the given concept and returns
   * its contents.
   *
   * @param concept The concept to get the implementation for.
   * @param form The concept form that the implementation must implement.
   * @param indent If true, returns indented implementation.
   * @return The first working implementation or an empty string if not
   *         applicable.
   */
  protected String getImplementation(Concept concept, String form,
    boolean indent)
  {
    Enumeration impls;
    Implementation impl;
    boolean implWorked = false;
    String result = new String();

    // try each concept implementation for this form until one works
    impls = concept.getImplementations(getTargetLanguage());
    while ((impls.hasMoreElements()) && (!implWorked))
    {
      impl = (Implementation) impls.nextElement();
      if (impl.getConceptForm().equals(form))
      { // implementations supports this concept form
        // try implementation
        try
        {
          result = getImplementation(impl);
          if (indent)
            result = indent(result);
          result += "\n";
          implWorked = true;
        }
        catch (CoComposeException e)
        { // try next
          if (log != null)
            log.log(e.getMessage());
        }
      }
    } //while
    return result;
  } //getImplementation

  /**
   * Checks and returns the implementation content, if check ok.
   * @param impl - the implementation to check
   * @return the content of impl, if check ok
   * @throws CoComposeException if check not ok
   */
  protected String getImplementation(Implementation impl) throws CoComposeException
  {
    impl.accept(this);
    if (log != null)
      log.log("Implementation worked");
    return impl.getContentWithoutTags();
  } //getImplementation

  /**
   * Writes the content to the designed file.
   *
   * @param dir The file directory.
   * @param filename The file name.
   * @param content The file content to write.
   * @throws CoComposeException if writing failed.
   */
  protected void writeFile(String dir, String filename, String content)
    throws CoComposeException
  {
    File outfile, outdir;
    OutputStream stream;
    PrintWriter out;
    int lfIndex, lastIndex = 0;

    try
    {
      outdir = new File(dir);
      if (!outdir.exists())
        outdir.mkdirs();
      outfile = new File(outdir, filename);
      if (outfile.exists())
        outfile.delete();
      outfile.createNewFile();
      // file is created, now write to it
      stream = new FileOutputStream(outfile);
      try
      {
        out = new PrintWriter(stream);
        try
        { // read lines and use println to output them
          while ((lfIndex = content.indexOf('\n', lastIndex)) != -1)
          {
            out.println(content.substring(lastIndex, lfIndex));
            lastIndex = lfIndex + 1;
          }
          out.print(content.substring(lastIndex));
        }
        finally
        {
          out.close();
        }
      }
      finally
      {
        stream.close();
      }
    }
    catch (Exception e)
    {
      throw new CoComposeException(e.getMessage());
    }
  } //writeFile

  /**
   * Dumps the implementation on standard out.
   *
   * @param model The model to dump the implementation for.
   * @param outdir The directory to save the dump file.
   * @throws CoComposeException if writing the implementation dump failed.
   */
  protected void dumpImplementation(Model model, String outdir)
    throws CoComposeException
  {
    StringBuffer impl = new StringBuffer();
    Hashtable implementations = getImpls();
    Enumeration entities = implementations.keys();
    Object entity;
    String entityName;

    if (log != null)
      log.log("Writing implementations to dump file");
    while (entities.hasMoreElements())
    {
      entity = entities.nextElement();
      // get name
      if (entity instanceof EDU.utwente.CoCompose.lang.AbstractConcept)
        entityName = ((AbstractConcept) entity).getName() + "(id = " +
          ((AbstractConcept) entity).getId() + ")";
      else if (entity instanceof EDU.utwente.CoCompose.util.UniqueRealm)
        entityName = ((UniqueRealm) entity).getName();
      else
        entityName = entity.toString();
      // write implementation block
      impl.append("-- start " + entityName + '\n');
      impl.append((String) implementations.get(entity) + '\n');
      impl.append("-- end " + entityName + '\n');
    } //while impls
    writeFile(outdir, model.getName() + ".dmp", impl.toString());
  } //dumpImplementation

  //////////////////////////////////////////////////////////////////////////////
  // Public methods
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Returns the target language this GenerateImplementationVisitor supports.
   *
   * @return The target language.
   */
  public abstract String getTargetLanguage();

  /**
   * Returns an indented implementation.
   *
   * @param impl The current implementation.
   * @return The indented implementation.
   */
  public String indent(String impl)
  {
    int lfIndex, lastIndex;
    String result;

    result = Indent;
    lfIndex = -1;
    lastIndex = 0;
    while ((lfIndex = impl.indexOf("\n", lfIndex + 1)) != -1)
    {
      result += impl.substring(lastIndex, lfIndex + 1) + Indent;
      lastIndex = lfIndex + 1;
    }
    result += impl.substring(lastIndex);
    return result;
  } //indent

  /**
   * Sets the implementation for a Concept.
   *
   * @param concept The concept to set the implementation for.
   * @param implementation The implementation to set.
   */
  public void setImpl(Concept concept, String implementation)
  {
    impls_.put(concept, implementation);
  } //setImpl(Concept concept, String implementation)

  /**
   * Sets the implementation for a UniqueRealm.
   *
   * @param realm The unique realm to set the implementation for.
   * @param implementation The implementation to set.
   */
  public void setImpl(UniqueRealm realm, String implementation)
  {
    impls_.put(realm, implementation);
  } //setImpl(Concept concept, String implementation)

  /**
   * Gets the implementation for a concept.
   *
   * @return The implementation.
   */
  public String getImpl(Concept concept)
  {
    if (impls_.containsKey(concept))
      return (String) impls_.get(concept);
    else
      return null;
  } //getImpl(Concept concept)

  /**
   * Gets the implementation for a unique realm.
   *
   * @return The implementation.
   */
  public String getImpl(UniqueRealm realm)
  {
    if (impls_.containsKey(realm))
      return (String) impls_.get(realm);
    else
      return null;
  } //getImpl(UniqueRealm realm)

  /**
   * Visits a Model object.
   *
   * @param model The model to visit.
   * @throws CoComposeException if visiting operation failed.
   */
  public void visitModel(Model model) throws CoComposeException
  {
    AbstractConcept concept;
    Iterator concepts;

    if (log != null)
    {
      log.log("Generating implementation for " + model.getName());
      log.incLogLevel();
    }
    try
    {
      // the model will now be enforced to be correct and complete
      model.checkCorrectness();
      model.checkCompleteness();
      // save forms of each concept
      concepts = model.getAbstractConcepts().iterator();
      while (concepts.hasNext())
      {
        concept = (AbstractConcept) concepts.next();
        if (concept instanceof Concept)
          setForm((Concept) concept, ((Concept) concept).getForm());
      }
      // create constraint checker visitor
      constraintCheck = new FormConstraintsVisitor();
      constraintCheck.setLogger(log);
      constraintCheck.setForms(getForms());
      // generate implementation for each concept
      concepts = model.getAbstractConcepts().iterator();
      while (concepts.hasNext())
      {
        concept = (AbstractConcept) concepts.next();
        if (concept instanceof Concept)
          ((Concept) concept).accept(this);
      }
      // apply composite implementation generators
      concepts = model.getAbstractConcepts().iterator();
      while (concepts.hasNext())
      {
        concept = (AbstractConcept) concepts.next();
        if (concept instanceof Composite)
          ((Composite) concept).accept(this);
      }
    }
    finally
    {
      if (log != null)
        log.decLogLevel();
    }
  } //visitModel

  /**
   * Generates implementation for a Concept object.
   *
   * @param concept The concept to visit.
   * @throws CoComposeException if generating implementation failed.
   */
  public void visitConcept(Concept concept) throws CoComposeException
  {
    String form = concept.getForm();

    // Check if concept form was determined
    if (form.length() < 1)
    {
      if (log != null)
        log.log("Concept form not determined for " + concept.getName());
      throw new CoComposeException("Concept form not determined for " +
        concept.getName(), concept);
    }
  } //visitConcept

  /**
   * Generates implementation for a Composite object.<BR>
   * PRE: The model attribute must be defined for the composite.
   *
   * @param composite The composite to visit.
   * @throws CoComposeException if generating implementation failed.
   */
  public void visitComposite(Composite composite) throws CoComposeException
  {
    Model model = composite.getModel();
    Enumeration implementations;
    ImplementationGenerator implementation;
    boolean implWorked = false;

    if (log != null)
    {
      log.log("Generating implementation for " + composite.getName());
      log.incLogLevel();
    }
    try
    {
      /* Precondition check -- disabled
      // check model
      if (model == null)
      {
        if (log != null)
          log.log("Error: " + composite.getName() + " has no model set");
        throw new CoComposeException(composite.getName() + " has no model set");
      }
      End precondition check */
      // try each implementation generator until successful
      implementations = composite.getImplementations(model.getTargetLanguage());
      while ((implementations.hasMoreElements()) && (!implWorked))
      {
        implementation = (ImplementationGenerator) implementations.nextElement();
        try
        {
          // apply implementation generator and set flag if it works
          implementation.accept(this);
          implWorked = true;
        }
        catch (CoComposeException e)
        { // try next
          if (log != null)
            log.log(e.getMessage());
        }
      }
      // throw exception if no implementation generator worked
      if (!implWorked)
      {
        if (log != null)
          log.log("No implementation generator worked for " + composite.getName());
        throw new CoComposeException("No implementation generator worked for " +
          composite.getName(), composite);
      }
    }
    finally
    {
      if (log != null)
        log.decLogLevel();
    }
  } //visitComposite

  /**
   * Generates implementation for an Implementation object.
   *
   * @param implementation The concept implementation to visit.
   * @throws CoComposeException if generating implementation failed.
   */
  public void visitImplementation(Implementation implementation)
    throws CoComposeException
  {
    if (log != null)
    {
      log.log("Applying concept implementation");
      log.incLogLevel();
    }
    try
    {
      // check this implementation
      if (log != null)
        log.log("Checking implementation constraints");
      implementation.accept(constraintCheck);
    }
    finally
    {
      if (log != null)
        log.decLogLevel();
    }
  } //visitImplementation

  /**
   * Generates implementation for an ImplementationGenerator object.
   *
   * @param implementation The implementation generator to visit.
   * @throws CoComposeException if generating implementation failed.
   */
  public void visitImplementationGenerator(ImplementationGenerator implementation)
    throws CoComposeException
  {
    File file;
    String filename;
    ImplementationGeneratorI impl;
    Composite composite = implementation.getComposite();

    if (log != null)
    {
      log.log("Generating implementation for implementation generator " +
        implementation.getFilename());
      log.incLogLevel();
    }
    try
    {
      // check this implementation generator
      if (log != null)
        log.log("Checking implementation generator constraints");
      implementation.accept(constraintCheck);

      // apply the implementation generator if check worked out
      if (log != null)
        log.log("Implementation generator constraints fulfilled, now applying");
      // load implementation generator
      file = new File(implementation.getFilename().replace('\\', '/'));
      filename = file.getName();
      filename = filename.substring(0, filename.length() - 4);
      try
      {
        impl = (ImplementationGeneratorI) Class.forName(
          IMPLGEN_PATH + filename).newInstance();
        if (log != null)
          log.log("Implementation generator plugin class " + IMPLGEN_PATH +
            filename + " loaded");
      }
      catch (Exception e)
      {
        if (log != null)
          log.log("Could not load implementation generator plugin class " +
            IMPLGEN_PATH + filename);
        throw new CoComposeException("Could not load implementation generator plugin class "
          + IMPLGEN_PATH + filename, composite);
      }
      try
      {
        impl.apply(implementation, this);
        if (log != null)
          log.log("Implementation generator plugin class " + IMPLGEN_PATH +
            filename + " applied");
      }
      catch (CoComposeException e)
      {
        if (log != null)
          log.log("Could not apply implementation generator plugin class " +
            IMPLGEN_PATH + filename + ": " + e.getMessage());
        throw new CoComposeException("Could not apply implementation generator plugin class " +
          IMPLGEN_PATH + filename + ": " + e.getMessage(), e.getConcept());
      }
      // find and apply role part for each role filling
    } //try
    finally
    {
      if (log != null)
        log.decLogLevel();
    }
  } //visitImplementationGenerator
} //class GenerateImplementationVisitor
