/**
 * Title:        CoCompose
 * Copyright:    Copyright (c) 2002, 2003
 * Author:       Dennis Wagelaar
 * Organisation: University of Twente, Vrije Universiteit Brussel
 * Description:  CoCompose Design Language Tool
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License, version 2 as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  (See the file "COPYING" that is included with this source distribution.)
 */

package EDU.utwente.CoCompose.translate;

import java.util.*;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeEvent;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.DefaultHandler;
import org.xml.sax.helpers.AttributesImpl;
import EDU.utwente.CoCompose.lang.*;
import EDU.utwente.CoCompose.util.CoComposeException;

/**
 * Represents an implementation for a Concept. It links to a file containing
 * the actual implementation, such as a Java source file.
 *
 * @author Dennis Wagelaar
 * @version 0.1.16, 31 October 2003
 */

public class Implementation implements PropertyChangeListener
{
  //////////////////////////////////////////////////////////////////////////////
  // Constants
  //////////////////////////////////////////////////////////////////////////////

  private static final String XMLNAME_IMPLEMENTATION = "Implementation";
  private static final String XMLNAME_CONTENT        = "ImplementationContent";
  private static final String XMLNAME_CONSTRAINTS    = "ImplementationConstraints";
  private static final String XMLNAME_CONSTRAINT     = "ImplementationConstraint";
  private static final String XMLNAME_FORM           = "ConceptForm";

  //////////////////////////////////////////////////////////////////////////////
  // Attributes
  //////////////////////////////////////////////////////////////////////////////

  private String language_ = new String();
  private String content_ = new String();
  private String conceptForm_ = new String();
  private Concept concept_ = null;
  // Constraints on concept form assumptions for concepts referred to in the
  // implementation. Structure: <AbstractConcept> |-> Collection(<ConceptForm>)
  private Hashtable constraints_ = new Hashtable();

  //////////////////////////////////////////////////////////////////////////////
  // Constructors
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Constructs an Implementation.
   */
  public Implementation()
  {
    super();
  } //Implementation

  //////////////////////////////////////////////////////////////////////////////
  // Protected methods
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Creates the constraints table XML using a SAX DefaultHandler.
   * @param handler the SAX DefaultHandler.
   * @throws SAXException - if an XML handler error occurs.
   */
  protected void constraintsToXML(DefaultHandler handler) throws SAXException
  {
    Hashtable constraints = getConstraints();
    AbstractConcept concept;
    Collection conceptForms;
    Iterator conceptFs;
    String form;
    AttributesImpl atts;

    // check for constraints
    if (constraints.size() < 1)
      return;

    // There are constraints, so create XML
    handler.startElement(null, null, XMLNAME_CONSTRAINTS, null);
    Enumeration constr = constraints.keys();
    while (constr.hasMoreElements())
    {
      concept = (AbstractConcept) constr.nextElement();
      atts = new AttributesImpl();
      atts.addAttribute(null, null, "Concept", "IDREF",
        String.valueOf(concept.getId()));
      handler.startElement(null, null, XMLNAME_CONSTRAINT, atts);
      conceptForms = (Collection) constraints.get(concept);
      conceptFs = conceptForms.iterator();
      while (conceptFs.hasNext())
      { // add each form
        form = (String) conceptFs.next();
        atts = new AttributesImpl();
        atts.addAttribute(null, null, "Form", "CDATA", form);
        handler.startElement(null, null, XMLNAME_FORM, atts);
        handler.endElement(null, null, XMLNAME_FORM);
      }
      handler.endElement(null, null, XMLNAME_CONSTRAINT);
    }
    handler.endElement(null, null, XMLNAME_CONSTRAINTS);
  } //constraintsToXML

  /**
   * Replaces names in source string and returns the new string. Replaces only
   * keywords, which are enclosed in &lt;&gt;.
   * @param source the source string.
   * @param nameTable the Hashtable&lt;String |-&gt; String&gt; of old names mapping to new names.
   * @param withoutTag if true, the new names will be put in without tag signs.
   * @return the new string.
   */
  protected String stringReplace(String source, Hashtable nameTable,
    boolean withoutTag)
  {
    Hashtable occurrences = new Hashtable();
    char c1, c2;
    Integer indexObject;
    int previndex, index;
    String oldname, newname, keyword;
    String result = new String();
    int srclength = source.length();
    SortedSet indexSet = new TreeSet();

    // find occurences first
    Enumeration names = nameTable.keys();
    while (names.hasMoreElements())
    {
      oldname = (String) names.nextElement();
      keyword = '<' + oldname + '>'; // form the corresponding keyword
      index = -1;
      while ((index = source.indexOf(keyword, index + 1)) >= 0)
      {
        // save index and what name occurred at this index
        indexObject = new Integer(index);
        occurrences.put(indexObject, oldname);
        indexSet.add(indexObject);
      } //while index
    } //while names
    // now replace occurences
    previndex = 0;
    Iterator indexes = indexSet.iterator();
    while (indexes.hasNext())
    {
      indexObject = (Integer) indexes.next();
      oldname = (String) occurrences.get(indexObject);
      newname = (String) nameTable.get(oldname);
      index = indexObject.intValue();
      if (index > previndex)
        result += source.substring(previndex, index);
      if (withoutTag)
        result += newname;
      else
        result += '<' + newname + '>';
      previndex = index + oldname.length() + 2; // keyword braces
    } //while names
    result += source.substring(previndex);
    return result;
  } //stringReplace

  /**
   * Unsubscribes from all AbstractConcepts on which we depend.
   */
  protected void unsubscribeAll()
  {
    AbstractConcept subject;

    Enumeration concepts = getConstraints().keys();
    while (concepts.hasMoreElements())
    {
      subject = (AbstractConcept) concepts.nextElement();
      subject.removePropertyChangeListener(this);
    }
  } //unsubscribeAll

  /**
   * Attempts to find the local equivalent/copy of the given concept in the
   * given model. This method checks what kind of AbstractConcept to look for
   * and uses appropriate methods to find it.
   * @param fromConcept the concept to look for
   * @param inModel the model to look in
   * @param instantiatedBy the local concept must be instantiated by this concept.
   * @return the local concept equivalent
   * @throws CoComposeException - if more than one concept dependency was found
   */
  protected AbstractConcept findLocalConcept(AbstractConcept fromConcept, 
    Model inModel, AbstractConcept instantiatedBy) throws CoComposeException
  {
    AbstractConcept localConcept;
    if (fromConcept instanceof EDU.utwente.CoCompose.lang.CompositePartConcept)
    { // find through composite
      AbstractConcept searchConcept = ((CompositePartConcept) fromConcept).getComposite();
      localConcept = searchLocalConcept(searchConcept, inModel, instantiatedBy);
      if (localConcept != null)
        localConcept = ((Composite) localConcept).getPartConcept(fromConcept.getName());
    }
    else
    {
      localConcept = searchLocalConcept(fromConcept, inModel, instantiatedBy);
    }
    if ((localConcept == null) &&
        (fromConcept instanceof EDU.utwente.CoCompose.lang.SolutionRole))
    { // retrieve the concept that fills this solution role
      localConcept = ((SolutionRole) fromConcept).getFilledBy();
    }
    if ((localConcept == null) && (inModel == fromConcept.getModel()))
    { // concept has not been copied, so this is the concept we need
      localConcept = fromConcept;
    }
    return localConcept;
  }

  /**
   * Searches the local equivalent/copy of the given concept in the
   * given model.
   * @param fromConcept the concept to look for
   * @param inModel the model to look in
   * @param instantiatedBy the local concept must be instantiated by this concept
   * @return the local concept equivalent or null
   * @throws CoComposeException - if more than one concept dependency was found
   */
  protected AbstractConcept searchLocalConcept(AbstractConcept fromConcept,
    Model inModel, AbstractConcept instantiatedBy) throws CoComposeException
  {
    AbstractConcept localConcept = null, thisConcept;
    Collection concepts = inModel.getAbstractConcepts(fromConcept);
    Iterator conceptsIt = concepts.iterator();
    int copyCount = 0;
    while (conceptsIt.hasNext())
    {
      thisConcept = (AbstractConcept) conceptsIt.next();
      // must be same instance of role part or default part (= null)
      if (thisConcept.getInstantiatedBy() != null)
      {
        if (fromConcept.getPartOf().equals(getConcept().getCopiedFrom().getPartOf()))
        { // same role part
          if (thisConcept.getInstantiatedBy().equals(instantiatedBy))
          {
            localConcept = thisConcept;
            copyCount++;
          }
        }
        else
        {
          localConcept = thisConcept;
          copyCount++;
        }
      }
      else
      {
        localConcept = thisConcept;
        copyCount++;
      }
      if (copyCount > 1)
      {
        throw new CoComposeException("More than one concept dependency" +
          " found in implementation for " + getConcept().getName());
      }
    }
    return localConcept;
  } //searchLocalConcept

  //////////////////////////////////////////////////////////////////////////////
  // Public methods
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Creates a constrained concept renaming event for this.
   * @param source the old constrained concept
   * @param target the new constrained concept
   * @return the renaming event
   */
  public static PropertyChangeEvent CreateRenameEvent(AbstractConcept source,
    AbstractConcept target)
  {
    return new PropertyChangeEvent(source, "Name", source.getPublicName(),
      target.getPublicName());
  } //CreateRenameEvent

  /**
   * Gets the language in which this implementation is expressed.
   * @return the implementation language.
   */
  public String getLanguage()
  {
    return language_;
  } //getLanguage

  /**
   * Sets the language in which this implementation is expressed.
   * @param language the implementation language.
   */
  public void setLanguage(String language)
  {
    language_ = language;
  } //setLanguage

  /**
   * Gets the implementation content.
   * @return the implementation content.
   */
  public String getContent()
  {
    return content_;
  } //getContent

  /**
   * Gets the implementation content without concept tags.
   * @return the implementation content without tag symbols.
   */
  public String getContentWithoutTags()
  {
    Concept concept;
    String content = getContent();
    Hashtable renameTable = new Hashtable();

    concept = getConcept();
    if (concept != null)
      renameTable.put(concept.getName(), concept.getName());
    Enumeration concepts = getConstraints().keys();
    while (concepts.hasMoreElements())
    {
      concept = (Concept) concepts.nextElement();
      renameTable.put(concept.getName(), concept.getName());
    }
    return stringReplace(content, renameTable, true);
  } //getContentWithoutTags

  /**
   * Sets the implementation content.
   * @param content the implementation content.
   */
  public void setContent(String content)
  {
    content_ = content;
  } //setContent

  /**
   * Gets the target language construct form for this implementation.
   * @return the target language construct form.
   */
  public String getConceptForm()
  {
    return conceptForm_;
  } //getConceptForm

  /**
   * Sets the target language construct form for this implementation.
   * @param conceptForm the target language construct form.
   */
  public void setConceptForm(String conceptForm)
  {
    conceptForm_ = conceptForm;
  } //setConceptForm

  /**
   * Gets the concept to which this implementation belongs.
   * @return the concept that owns this implementation.
   */
  public Concept getConcept()
  {
    return concept_;
  } //getConcept

  /**
   * Sets the concept to which this implementation belongs.
   * @param concept the concept that owns this implementation.
   */
  public void setConcept(Concept concept)
  {
    concept_ = concept;
  } //setConcept

  /**
   * Adds a constraint for a concept referred to in this implementation.
   * Semantics: &lt;concept&gt; |-&gt; (&lt;conceptForm&gt; | &lt;conceptForm&gt; in &lt;conceptForms&gt;)
   * @param concept the concept to which the constraints refers.
   * @param conceptForms the Collection&lt;String&gt; of concepts form of which one is assumed
   *        for the given concept (the actual constraint).
   */
  public void addConstraint(AbstractConcept concept, Collection conceptForms)
  {
    if (!conceptForms.isEmpty())
    {
      constraints_.put(concept, conceptForms);
      concept.addPropertyChangeListener(this);
    }
  } //addConstraint

  /**
   * Removes the constraint referring to the given concept.
   * @param concept the concept for which to remove the constraint.
   */
  public void removeConstraint(AbstractConcept concept)
  {
    concept.removePropertyChangeListener(this);
    constraints_.remove(concept);
  } //removeConstraint

  /**
   * Clears all constraints.
   */
  public void clearConstraints()
  {
    unsubscribeAll();
    constraints_.clear();
  } //clearConstraints

  /**
   * Gets the concept form constraints table.
   * @return a copy of the constraints table.
   *         Structure: Hashtable&lt;AbstractConcept |-&gt; Collection&lt;conceptForm:String&gt;&gt;
   */
  public Hashtable getConstraints()
  {
    return constraints_;
  } //getConstraints

  /**
   * Gets the model for this implementation from the concept.
   * @return the model for this implementation's concept or null.
   */
  public Model getModel()
  {
    Concept concept = getConcept();
    if (concept != null)
      return concept.getModel();
    else
      return null;
  } //getModel

  /**
   * Copies the attributes from the given implementation to self. It doesn't
   * copy the concept link or concept constraints.
   * @param fromImplementation the implementation to copy from.
   */
  public void copyAttributes(Implementation fromImplementation)
  {
    setLanguage(fromImplementation.getLanguage());
    setConceptForm(fromImplementation.getConceptForm());
    setContent(fromImplementation.getContent());
  } //copyAttributes

  /**
   * Copies the constraints from the given implementation.
   * @param fromImplementation the implementation to copy from.
   * @throws CoComposeException - if a source concept for a constraint could not
   *         be found in the current model.
   */
  public void copyConstraints(Implementation fromImplementation)
    throws CoComposeException
  {
    AbstractConcept concept, localConcept;
    Collection forms, newForms;
    Model model = getModel();
    Hashtable constraints = fromImplementation.getConstraints();

    // copy constraints
    clearConstraints();
    Enumeration constr = constraints.keys();
    while (constr.hasMoreElements())
    {
      concept = (AbstractConcept) constr.nextElement();
      forms = (Collection) constraints.get(concept);
      newForms = new Vector(forms);
      if (model == null)
      { // assume this concept resides in same model
        localConcept = concept;
      }
      else
      { // make sure the concept exists in the current model by retrieving it
        // through the model
        localConcept = findLocalConcept(concept, model, getConcept().getInstantiatedBy());
      }
      if (localConcept != null)
      {
        addConstraint(localConcept, newForms);
      }
      else
      {
        CoComposeException e = new CoComposeException("Could not find " + concept.getName() +
          " in the current model", concept);
        e.printStackTrace();
        throw e;
      }
      if (!localConcept.getPublicName().equals(concept.getPublicName()))
      { // adapt implementation content
        propertyChange(CreateRenameEvent(concept, localConcept));
      }
    }
  } //copyConstraints

  /**
   * Create XML from this object using a SAX DefaultHandler.
   * @param handler the SAX DefaultHandler.
   * @throws SAXException - if an XML handler error occurs.
   */
  public void toXML(DefaultHandler handler) throws SAXException
  {
    // Create attributes
    AttributesImpl atts = new AttributesImpl();
    atts.addAttribute(null, null, "Language", "CDATA", getLanguage());
    atts.addAttribute(null, null, "ConceptForm", "CDATA", getConceptForm());

    // Write XML
    handler.startElement(null, null, XMLNAME_IMPLEMENTATION, atts);

    // Implementation content
    handler.startElement(null, null, XMLNAME_CONTENT, null);
    String content = getContent();
    handler.characters(content.toCharArray(), 0, content.length());
    handler.endElement(null, null, XMLNAME_CONTENT);

    // Constraints
    constraintsToXML(handler);

    handler.endElement(null, null, XMLNAME_IMPLEMENTATION);
  } //toXML

  /**
   * Adapts the concept data used in the implementation (content & constraints).
   * @param adapted A table of the form: Hashtable&lt;oldConcept:Concept |-&gt; newConcept:Concept&gt;
   */
  public void adapt(Hashtable adapted)
  {
    Hashtable constraints = getConstraints();
    String name;
    Collection forms;
    AbstractConcept oldConcept, newConcept;
    Hashtable localTable = new Hashtable();

    // adapt constraints
    Enumeration oldConcepts = adapted.keys();
    while (oldConcepts.hasMoreElements())
    {
      oldConcept = (AbstractConcept) oldConcepts.nextElement();
      newConcept = (AbstractConcept) adapted.get(oldConcept);
      if (constraints.containsKey(oldConcept))
      { // replace constraint
        forms = (Collection) constraints.get(oldConcept);
        removeConstraint(oldConcept);
        addConstraint(newConcept, forms);
      }
      if (constraints.containsKey(newConcept))
      { // add this renameTable entry to the local table
        localTable.put(oldConcept.getName(),
          ((AbstractConcept) adapted.get(oldConcept)).getName());
      }
    } //while
    // adapt implementation content
    String content = getContent();
    content = stringReplace(content, localTable, false);
    setContent(content);
  } //adapt

  /**
   * This method gets called when a property is changed.
   * @param evt the event describing which property from what object has been
   *            changed.
   */
  public void propertyChange(PropertyChangeEvent evt)
  {
    AbstractConcept src;
    Hashtable renameTable;

    // only listen to AbstractConcept objects
    src = (AbstractConcept) evt.getSource();
    if (!(src instanceof EDU.utwente.CoCompose.lang.AbstractConcept))
      return;
    if (evt.getPropertyName().equals("Name"))
    {
      // create renaming table
      renameTable = new Hashtable();
      renameTable.put(evt.getOldValue(), evt.getNewValue());
      // adapt implementation content
      String content = getContent();
      content = stringReplace(content, renameTable, false);
      setContent(content);
    }
    else if (evt.getPropertyName().equals("Removed"))
    { // this concept has been removed from the model, so stop listening
      removeConstraint(src);
    }
  } //propertyChange

  /**
   * Accepts a ModelVisitor.
   * @param visitor the visitor object to accept.
   * @throws CoComposeException - if visiting operation failed.
   */
  public void accept(ModelVisitor visitor) throws CoComposeException
  {
    visitor.visitImplementation(this);
  } //accept
} //class Implementation
